"""
Agnocomplete constants
"""

"Agnocomplete default page size"
AGNOCOMPLETE_DEFAULT_PAGESIZE = 10

"Agnocomplete maximum page size"
AGNOCOMPLETE_MAX_PAGESIZE = 100

"Agnocomplete minimum page size"
AGNOCOMPLETE_MIN_PAGESIZE = 3

"Agnocomplete default query size"
AGNOCOMPLETE_DEFAULT_QUERYSIZE = 3

"Agnocomplete minimum query size"
AGNOCOMPLETE_MIN_QUERYSIZE = 2

"Default data attribute name for fields"
AGNOCOMPLETE_DATA_ATTRIBUTE = 'agnocomplete'

"The context attribute name given to fields"
AGNOCOMPLETE_USER_ATTRIBUTE = '__user'
